#include <windows.h>
#include <math.h>
#include "lsmtypes.h"
#include "calc.h"

float   frac(float a) {
        return a-floorf(a);
}

bool isPowerof2(unsigned int value) {
    while(value>1) {
        if(value&1) return false;
        value>>=1;
    }

    return true;
}

unsigned int nearestPowerof2(unsigned int value) {
    unsigned int pow;

    value=(value<<8)/181;

    pow=1;
    while (value>1) {
        value>>=1;
        pow<<=1;
    }
    return pow;
}

float errorCorrection(float raw) {
        float tmp;

        tmp=(float)floor(raw*16384+.5);
        if (tmp<16384.0 && tmp>=0.0) tmp=16384.0;
        if (tmp>-16384.0 && tmp<0.0) tmp=-16384.0;
        raw+=tmp;
        return raw-tmp;
}

#if(0)
// This function calculates the point at which a given ray will intersect a
// given cylinder.  If the ray is parallel to the cylinder's axis, or if the
// ray misses the cylinder, result is left untouched.

bool traceCylinder(Vec3& pc,Vec3& va,float r,Vec3& pr,Vec3& vr,Vec3& result) {
    float dv=va.dot(vr);
    float ca=1.0f-dv*dv;
    if(ca==0.0f) return false;

    Vec3 diff;
    (diff=pc)-=pr;
    float dd=va.dot(diff);
    float cb=diff.dot(vr)-dd*dv;
    float cc=diff.dot(diff)-dd*dd-r*r;

    if(ca*cc>cb*cb) return false;

    float k=(cb-sqrt(cb*cb-ca*cc))/ca;

    ((result=vr)*=k)+=pr;
    return true;
}

// This function takes a point and moves it to the nearest point on a given plane
// This is always possible (no chance of divide-by-zero or square-root of
// negative value) so there is no need to report on success.
// return value is not build in result so that user can pass same reference as
// test point and result point

void nearest2Plane(Vec3& pc,Vec3& vn,Vec3& pt,Vec3& result) {
    Vec3 diff;

    (diff=pc)-=pt;
    float k=diff.dot(vn);

    ((diff=vn)*=k)+=pt;
    result=diff;
}    
#endif

// This function takes a vertex (start) and moves it along the lsmView's forward
// vector so that it lies in the plane, normal to norm and passing through pp;
// used to confine a vertex to a plane but allow it to appear to move with the
// mouse.

bool    tracePlane(Vec3& pp,Vec3& vn,Vec3& pr,Vec3& vr,Vec3& result) {
    float div=vr.dot(vn);
    if(fabs(div)<.01) return false; // does not meaningfully intersect

    Vec3 diff;
    (diff=pp)-=pr;

    float dist=diff.dot(vn)/div;

    ((diff=vr)*=dist)+=pr;

    result=diff;
    return true;
}

// This function takes a ray (ps+k*vp), and finds on it the point that is
// nearest to the line (mp+k*vp).

bool nearest2Line(Vec3& ps,Vec3& vs,Vec3& mp,Vec3& vp,Vec3& result) {
    Vec3 temp;
    Vec3 norm;

    temp.cross(vs,vp);
    norm.cross(temp,vp);
    norm.normalize();
    return tracePlane(mp,norm,ps,vs,result);
}

// traces a ray through a sphere
// resolves to a point on the perimeter of the sphere if the ray doesn't intersect
bool traceSphere(Vec3& pc,float rad,Vec3& pr,Vec3& vr,Vec3& result) {
    Vec3 pd;
    float cb,cc,k;
    bool flag;

    (pd=pc)-=pr;

    cb=vr.dot(pd);
    cc=pd.dot(pd)-rad*rad;

    k=cb;
    flag=(cb*cb>cc);
    if(flag) k-=sqrtf(cb*cb-cc);

    ((pd=vr)*=k)+=pr;

    if(!flag) {
      pd-=pc;
      pd.normalize();
      (pd*=rad)+=pc;
    }
    result=pd;
    return flag;
}

// stuff for the Vec2 and Vec3 objects

Vec3::Vec3() : x(0.0f), y(0.0f),z(0.0f) { }

Vec3::Vec3(float a) : x(a), y(a), z(a) { }

//Vec3::Vec3(float* a) : x(a[0]), y(a[1]), z(a[2]) { }

bool Vec3::operator!=(Vec3& ov) {
    if(x!=ov.x) return true;
    if(y!=ov.y) return true;
    return (z!=ov.z);
}

float Vec3::dot(Vec3& ov) {
    return x*ov.x + y*ov.y + z*ov.z;
}

float Vec3::dot(float* ov) {
    return x*ov[0] + y*ov[1] + z*ov[2];
}

void Vec3::set(float a) {
    x=a;y=a;z=a;
}

Vec2& Vec2::operator*=(float a) {
    u*=a;
    v*=a;
    return *this;
}

Vec3& Vec3::operator*=(float a) {
    x*=a;
    y*=a;
    z*=a;
    return *this;
}

Vec2& Vec2::operator/=(float a) {
     u/=a;
     v/=a;
     return *this;
}

Vec3& Vec3::operator/=(float a) {
    x/=a;
    y/=a;
    z/=a;
    return *this;
}

Vec2& Vec2::operator+=(Vec2& ov) {
    u+=ov.u;
    v+=ov.v;
    return *this;
}

Vec3& Vec3::operator+=(Vec3& ov) {
    x+=ov.x;
    y+=ov.y;
    z+=ov.z;
    return *this;
}

Vec2& Vec2::operator-=(Vec2& ov) {
    u-=ov.u;
    v-=ov.v;
    return *this;
}

Vec3& Vec3::operator-=(Vec3& ov) {
    x-=ov.x;
    y-=ov.y;
    z-=ov.z;
    return *this;
}

Vec3& Vec3::cross(Vec3& a,Vec3& b) {
    x=a.y*b.z - a.z*b.y;
    y=a.z*b.x - a.x*b.z;
    z=a.x*b.y - a.y*b.x;
    return *this;
}

float Vec3::normalize() {
    return normalize(*this);
}

float Vec3::normalize(Vec3& s) {
    float l;

    l=sqrtf(s.dot(s));
    *this=s;
    if(l!=0.0) *this/=l;
    return l;
}

float Vec3::distance2(Vec3& ov) {
    return (x-ov.x)*(x-ov.x) + (y-ov.y)*(y-ov.y) + (z-ov.z)*(z-ov.z);
}

float Vec3::distance(Vec3& ov) {
    return sqrtf(distance2(ov));
}

float Vec3::length() {
    return sqrtf(dot(*this));
}

void Vec2::put(float* d) {
    if(d) {
        d[0]=u;
        d[1]=v;
    }
}

void Vec3::put(float* d) {
    if(d) {
        d[0]=x;
        d[1]=y;
        d[2]=z;
    }
}

