// undo.c -- implements all of the undo and redo functions for the Model class
// #include "userint.h"
#include "model.h"
#include "scene.h"
#include "undo.h"
#include "funcs.h"
#include "vert.h"
#include "edge.h"
#include "face.h"
#include "uvpt.h"
#include "bone.h"
#include "mouse.h"

void    Model::destroyAllHistory() {
    MYPTR   m;

    mergeDeleted();

// delete all historical data
    STEPBONES(this,m.b) m.b->destroyHistory();
    STEPUVPTS(this,m.u) m.u->destroyHistory();
    STEPTINTS(this,m.t) m.t->destroyHistory();
    STEPVERTS(this,m.v) m.v->destroyHistory();
    STEPEDGES(this,m.e) m.e->destroyHistory();
    STEPFACES(this,m.f) m.f->destroyHistory();

    currentVersion=0;
    highVersion=0;

    separateDeleted();
}

void    Scene::destroyAllHistory() { }

void Model::takeSnap() {
        MYPTR m;
        bool    update;

//      ReportSystemError("takeSnap() entered.");
        if(modelFlags&MF_REORIENT) orientWindings();

//      ReportSystemError("Returned from orientWindings().");

        currentVersion++;

        mergeDeleted();

//      ReportSystemError("Returned from mergeDeleted().");

        update=false;

        STEPBONES(this,m.b) update|=m.b->takeSnap(currentVersion);
//      ReportSystemError("STEPBONES() done.");
        STEPUVPTS(this,m.u) update|=m.u->takeSnap(currentVersion);
//      ReportSystemError("STEPUVPTS() done.");
        STEPTINTS(this,m.t) update|=m.t->takeSnap(currentVersion);
//      ReportSystemError("STEPTINTS() done.");
        STEPVERTS(this,m.v) update|=m.v->takeSnap(currentVersion);
//      ReportSystemError("STEPVERTS() done.");
        STEPEDGES(this,m.e) update|=m.e->takeSnap(currentVersion);
//      ReportSystemError("STEPEDGES() done.");
        STEPFACES(this,m.f) update|=m.f->takeSnap(currentVersion);
//      ReportSystemError("STEPFACES() done.");

        separateDeleted();
//      ReportSystemError("Returned from separateDeleted().");

//        validate();
//      ReportSystemError("Returned from validate().");

        if(currentVersion>0 && update==false) currentVersion--;
        highVersion=currentVersion;

        updateToolBar();
//      ReportSystemError("Returned from UpdateToolBar().");
}

void Model::undoChanges() {
        MYPTR m;

        if(currentVersion==0) return;

        currentVersion--;

        mergeDeleted();

//      ReportSystemError("About to call restoreVersion()");

        STEPBONES(this,m.b) m.b->restoreVersion(currentVersion);
        STEPUVPTS(this,m.u) m.u->restoreVersion(currentVersion);
        STEPTINTS(this,m.t) m.t->restoreVersion(currentVersion);
        STEPVERTS(this,m.v) m.v->restoreVersion(currentVersion);
        STEPEDGES(this,m.e) m.e->restoreVersion(currentVersion);
        STEPFACES(this,m.f) m.f->restoreVersion(currentVersion);

//      ReportSystemError("Done with that.");

        clearAll();
        clearHighlight();
        setMouseMode(MM_NOTHING);

        separateDeleted();

        validateMapping();
        validate();

        updateSkeleton(false);

        modelFlags|=MF_BUILD_SUBD;//|MF_REBUILD_NORMS;

        updateToolBar();
}

void Model::redoChanges() {
        MYPTR m;

//        UserInt::updateStatusBar("Starting redo operation",0,0);

        if(currentVersion==highVersion) return;

        currentVersion++;

        mergeDeleted();

        STEPBONES(this,m.b) m.b->restoreVersion(currentVersion);
        STEPUVPTS(this,m.u) m.u->restoreVersion(currentVersion);
        STEPTINTS(this,m.t) m.t->restoreVersion(currentVersion);
        STEPVERTS(this,m.v) m.v->restoreVersion(currentVersion);
        STEPEDGES(this,m.e) m.e->restoreVersion(currentVersion);
        STEPFACES(this,m.f) m.f->restoreVersion(currentVersion);

        clearAll();
        clearHighlight();
        setMouseMode(MM_NOTHING);

        separateDeleted();

        validateMapping();
        validate();

        updateSkeleton(false);

        modelFlags|=MF_BUILD_SUBD;//|MF_REBUILD_NORMS;

        updateToolBar();
}

void Model::mergeDeleted() {

        while(delBones.first()) {
                Bone* b;
                b=(Bone*)(delBones.first());
                b->remove();
                Bones.addTail(b);
        }

        while(delTints.first()) {
                Tint* t;
                t=(Tint*)(delTints.first());
                t->remove();
                Tints.addTail(t);
        }

        while(delUVPts.first()) {
                UVPt* u;
                u=(UVPt*)(delUVPts.first());
                u->remove();
                UVPts.addTail(u);
        }

        while(delVerts.first()) {
                Vert* v;
                v=(Vert*)(delVerts.first());
                v->remove();
                Vertices.addTail(v);
        }

        while(delEdges.first()) {
                Edge* e;
                e=(Edge*)(delEdges.first());
                e->remove();
                Edges.addTail(e);
        }

        while(delFaces.first()) {
                Face* f;
                f=(Face*)(delFaces.first());
                f->remove();
                Faces.addTail(f);
        }
}

void Model::separateDeleted() {
        MYPTR   m;
        MYPTR   nm;

        m.f=FIRSTFACE(this);
        while(m.f) {
                nm.f=(Face*)(m.f->next());
                if(m.f->flags&TF_KILLME) {
                        nukeFace(m.f);
                } else {
                        if(m.f->flags&TF_INVALID) {
                                m.f->remove();
                                delFaces.addTail(m.f);
                        }
                }
                m.f=nm.f;
        }

        m.e=FIRSTEDGE(this);
        while(m.e) {
                nm.e=(Edge*)(m.e->next());
                if(m.e->flags&TF_KILLME) {
                        nukeEdge(m.e);
                } else {
                        if(m.e->flags&TF_INVALID) {
                                m.e->remove();
                                delEdges.addTail(m.e);
                        }
                }
                m.e=nm.e;
        }

        m.v=FIRSTVERT(this);
        while(m.v) {
                nm.v=(Vert*)(m.v->next());
                if(m.v->flags&TF_KILLME) {
                        nukeVertex(m.v);
                } else {
                        if(m.v->flags&TF_INVALID) {
                                m.v->remove();
                                delVerts.addTail(m.v);
                        }
                }
                m.v=nm.v;
        }

        m.b=FIRSTBONE(this);
        while(m.b) {
                nm.b=(Bone*)(m.b->next());
                if(m.b->flags&TF_KILLME) {
                        nukeBone(m.b);
                } else {
                        if(m.b->flags&TF_INVALID) {
                                m.b->remove();
                                delBones.addTail(m.b);
                        }
                }
                m.b=nm.b;
        }

        m.u=FIRSTUVPT(this);
        while(m.u) {
                nm.u=(UVPt*)(m.u->next());
                if(m.u->flags&TF_KILLME) {
                        nukeUVPt(m.u);
                } else {
                        if(m.u->flags&TF_INVALID) {
                                m.u->remove();
                                delUVPts.addTail(m.u);
                        }
                }
                m.u=nm.u;
        }

        m.t=FIRSTTINT(this);
        while(m.t) {
                nm.t=(Tint*)(m.t->next());
                if(m.t->flags&TF_KILLME) {
                        nukeTint(m.t);
                } else {
                        if(m.t->flags&TF_INVALID) {
                                m.t->remove();
                                delTints.addTail(m.t);
                        }
                }
                m.t=nm.t;
        }
}

